@@ -2,13 +2,16 @@
 Defines parsing functions and the `Parser` class that they instantiate.
 """
 # pyright: reportGeneralTypeIssues=false
+from __future__ import annotations
+
 import operator
 import os
+import re
 import sys
 import typing
 from dataclasses import asdict, dataclass, replace
 from functools import lru_cache, partial, reduce
-from typing import Any, Callable, Dict, Generator, Generic, Optional, Type, TypeVar
+from typing import Any, Callable, Dict, Generic, Optional, Type, TypeVar
 
 from pytypeclass import Monad, MonadPlus
 from pytypeclass.nonempty_list import NonemptyList
@@ -262,11 +265,11 @@ class Parser(MonadPlus[A_co]):
         Let's start with our simplest parser, `argument`:
         >>> p1 = argument("some_dest")
 
-        Now let's use the `equals` parser to write a function that takes the output of `p1` and fails unless
+        Now let's use the `matches` parser to write a function that takes the output of `p1` and fails unless
         the next argument is the same as the first:
-        >>> def f(kvs: Sequence(KeyValue[str])) -> Parser[Sequence[KeyValue[str]]]:
+        >>> def f(kvs: Sequence[KeyValue[str]]) -> Parser[Sequence[KeyValue[str]]]:
         ...     [kv] = kvs
-        ...     return equals(kv.value)
+        ...     return matches(kv.value)
 
         >>> p = p1 >= f
         >>> p.parse_args("a", "a")
@@ -335,7 +338,34 @@ class Parser(MonadPlus[A_co]):
             if error.usage:
                 print(error.usage)
 
-    def many(self: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+    def ignore(self: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+        """
+        Ignores the output from a parser. This is useful when you expect
+        to give arguments to the command line that some other utility will
+        handle.
+
+        >>> p = flag("hello").ignore()
+
+        This will not bind any value to `"hello"`:
+        >>> p.parse_args("--hello")
+        {}
+
+        But `--hello` is still required:
+        >>> p.parse_args()
+        The following arguments are required: --hello
+        """
+
+        def g(keep: Parse[Sequence[A]]) -> Result[Parse[Sequence[A]]]:
+            return Result(NonemptyList(Parse(Sequence([]), keep.unparsed)))
+
+        def f(cs: Sequence[str]) -> Result[Parse[Sequence[A]]]:
+            return self.parse(cs) >= g
+
+        return Parser(f, usage=None, helps={})
+
+    def many(
+        self: "Parser[Sequence[A]]", max: Optional[int] = None
+    ) -> "Parser[Sequence[A]]":
         """
         Applies `self` zero or more times (like `*` in regexes).
 
@@ -357,10 +387,18 @@ class Parser(MonadPlus[A_co]):
         >>> p.parse_args("--verbose", "--quiet", return_dict=False) # mix --verbose and --quiet
         [('verbose', True), ('quiet', True)]
         """
-        p = self.many1() | self.empty()
+        if max == 0:
+            p = self.empty()
+        else:
+            if max is not None:
+                max -= 1
+                assert max >= 0, max
+            p = self.many1(max=max) | self.empty()
         return replace(p, usage=f"[{self.usage} ...]")
 
-    def many1(self: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+    def many1(
+        self: "Parser[Sequence[A]]", max: Optional[int] = None
+    ) -> "Parser[Sequence[A]]":
         """
         Applies `self` one or more times (like `+` in regexes).
 
@@ -375,17 +413,9 @@ class Parser(MonadPlus[A_co]):
         The following arguments are required: 1-or-more
         """
 
-        def g() -> Generator["Parser[Sequence[A]]", Sequence[A], None]:
-            # noinspection PyTypeChecker
-            r1: Sequence[A] = yield self
-            # noinspection PyTypeChecker
-            r2: Sequence[A] = yield self.many()
-            yield Parser[Sequence[A]].return_(r1 + r2)
-
         @lru_cache()
         def f(cs: tuple):
-            y = Parser.do(g)
-            assert isinstance(y, Parser), y
+            y = self >> self.many(max=max)
             return y.parse(Sequence(list(cs)))
 
         return Parser(
@@ -581,8 +611,8 @@ class Parser(MonadPlus[A_co]):
         help: Print this help message.
 
         We can use `wrap_help` to print partial usage messages, e.g. for subcommands:
-        >>> subcommand1 = equals("subcommand1") >> option("option1").wrap_help()
-        >>> subcommand2 = equals("subcommand2") >> option("option2").wrap_help()
+        >>> subcommand1 = matches("subcommand1") >> option("option1").wrap_help()
+        >>> subcommand2 = matches("subcommand2") >> option("option2").wrap_help()
         >>> p = subcommand1 | subcommand2
         >>> p.parse_args("subcommand1", "-h")
         usage: --option1 OPTION1
@@ -734,13 +764,15 @@ def done() -> Parser[Sequence[A]]:
     return Parser(f, usage=None, helps={})
 
 
-def equals(s: str, peak=False) -> Parser[Sequence[KeyValue[str]]]:
+def matches(
+    s: str, peak: bool = False, regex: bool = True
+) -> Parser[Sequence[KeyValue[str]]]:
     """
     Checks if the next word is `s`.
 
-    >>> equals("hello").parse_args("hello")
+    >>> matches("hello").parse_args("hello")
     {'hello': 'hello'}
-    >>> equals("hello").parse_args("goodbye")
+    >>> matches("hello").parse_args("goodbye")
     usage: hello
     Expected 'hello'. Got 'goodbye'
 
@@ -752,28 +784,39 @@ def equals(s: str, peak=False) -> Parser[Sequence[KeyValue[str]]]:
         If `False`, then the parser will consume the word and return the remaining words as `unparsed`.
         If `True`, then the parser leaves the `unparsed` component unchanged.
 
+    regex : bool
+        Whether to treat `s` as a regular expression. If `False`, then the parser will only succeed on
+        string equality.
+
     Examples
     --------
 
-    >>> p = equals("hello") >> equals("goodbye")
+    >>> p = matches("hello") >> matches("goodbye")
     >>> p.parse_args("hello", "goodbye")
     {'hello': 'hello', 'goodbye': 'goodbye'}
 
     Look what happens when `peak=True`:
-    >>> p = equals("hello", peak=True) >> equals("goodbye")
+    >>> p = matches("hello", peak=True) >> matches("goodbye")
     >>> p.parse_args("hello", "goodbye")
     usage: hello goodbye
     Expected 'goodbye'. Got 'hello'
 
     The first parser didn't consume the word and so "hello" got passed on to `equals("goodbye")`.
     But this would work:
-    >>> p = equals("hello", peak=True) >> equals("hello") >>equals("goodbye")
+    >>> p = matches("hello", peak=True) >> matches("hello") >>matches("goodbye")
     >>> p.parse_args("hello", "goodbye")
     {'hello': 'hello', 'goodbye': 'goodbye'}
     """
+
+    def predicate(_s: str) -> bool:
+        if regex:
+            return bool(re.match(s, _s))
+        else:
+            return s == _s
+
     if peak:
         return sat_peak(
-            predicate=lambda _s: _s == s,
+            predicate=predicate,
             on_fail=lambda _s: UnequalError(
                 left=s, right=_s, usage=f"Expected '{s}'. Got '{_s}'"
             ),
@@ -781,7 +824,7 @@ def equals(s: str, peak=False) -> Parser[Sequence[KeyValue[str]]]:
         )
     else:
         return sat(
-            predicate=lambda _s: _s == s,
+            predicate=predicate,
             on_fail=lambda _s: UnequalError(
                 left=s, right=_s, usage=f"Expected '{s}'. Got '{_s}'"
             ),
@@ -866,7 +909,7 @@ def flag(
         cs: Sequence[str],
         s: str,
     ) -> Result[Parse[Sequence[KeyValue[bool]]]]:
-        parser = equals(s) >= (lambda _: defaults(**{dest: not default}))
+        parser = matches(s) >= (lambda _: defaults(**{dest: not default}))
         return parser.parse(cs)
 
     parser = Parser(partial(f, s=_string), usage=None, helps={})
@@ -885,7 +928,7 @@ def help_parser(usage: Optional[str], parsed: A) -> Parser[A]:
     def f(
         cs: Sequence[str],
     ) -> Result[Parse[A]]:
-        result = (equals("--help", peak=True) | equals("-h", peak=True)).parse(cs)
+        result = (matches("--help", peak=True) | matches("-h", peak=True)).parse(cs)
         if isinstance(result.get, ArgumentError):
             return Result.return_(Parse(parsed=parsed, unparsed=cs))
         return Result(HelpError(usage=usage or "Usage not provided."))
@@ -940,7 +983,9 @@ def item(
     return Parser(f, usage=name, helps={})
 
 
-def nonpositional(*parsers: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
+def nonpositional(
+    *parsers: "Parser[Sequence[A]]", repeated: Optional[Parser[Sequence[A]]] = None
+) -> "Parser[Sequence[A]]":
     """
     `nonpositional` takes a sequence of parsers as arguments and attempts all permutations of them,
     returning the first permutations that is successful:
@@ -950,18 +995,71 @@ def nonpositional(*parsers: "Parser[Sequence[A]]") -> "Parser[Sequence[A]]":
     {'verbose': True, 'quiet': True}
     >>> p.parse_args("--quiet", "--verbose")  # reverse order also works
     {'quiet': True, 'verbose': True}
+
+    Parameters
+    ----------
+    repeated : Optional[Parser[Sequence[A]]]
+        If provided, this parser gets applied repeatedly (zero or more times) at all positions.
+
+    Examples
+    --------
+    >>> p = nonpositional(repeated=flag("x"))
+    >>> p.parse_args(return_dict=False)
+    []
+    >>> p.parse_args("-x", return_dict=False)
+    [('x', True)]
+    >>> p.parse_args("-x", "-x", return_dict=False)
+    [('x', True), ('x', True)]
+
+    >>> p = nonpositional(flag("y"), repeated=flag("x"))
+    >>> p.parse_args("-y", return_dict=False)
+    [('y', True)]
+    >>> p.parse_args("-y", "-x", return_dict=False)
+    [('y', True), ('x', True)]
+    >>> p.parse_args("-x", "-y", return_dict=False)
+    [('x', True), ('y', True)]
+    >>> p.parse_args("-y", "-x", "-x", return_dict=False)
+    [('y', True), ('x', True), ('x', True)]
+    >>> p.parse_args("-x", "-y", "-x", return_dict=False)
+    [('x', True), ('y', True), ('x', True)]
+    >>> p.parse_args("-x", "-x", "-y", return_dict=False)
+    [('x', True), ('x', True), ('y', True)]
+
+    >>> p = nonpositional(flag("y"), repeated=(flag("x") | flag("z")).ignore())
+    >>> p.parse_args("-x", "-y", "-z", return_dict=False)
+    [('y', True)]
     """
+    sep = " " if len(parsers) <= 3 else "\n"
+    _parsers = [*parsers] if repeated is None else [*parsers, repeated]
+    usage = sep.join([p.usage or "" for p in _parsers])
+    repeat_parser = None
+    if repeated is not None:
+        _repeated = repeated  # for mypy's benefit
+
+        def f(cs: Sequence[str]):
+            p = _repeated >> nonpositional(*parsers, repeated=repeated)
+            return p.parse(cs)
+
+        repeat_parser = Parser(f, usage=None, helps={})
+
     if not parsers:
-        return empty()
+        if repeat_parser is None:
+            return empty()
+        else:
+            return repeat_parser | empty()
 
     def get_alternatives():
+        if repeat_parser is not None:
+            yield repeat_parser
         for i, head in enumerate(parsers):
             tail = [p for j, p in enumerate(parsers) if j != i]
-            yield head >> nonpositional(*tail)
+            yield head >> nonpositional(*tail, repeated=repeated)
 
     parser = reduce(operator.or_, get_alternatives())
-    sep = " " if len(parsers) <= 3 else "\n"
-    return replace(parser, usage=sep.join([p.usage or "" for p in parsers]))
+    helps = parser.helps
+    if repeated is not None:
+        helps = {**helps, **repeated.helps}
+    return replace(parser, usage=usage, helps=helps)
 
 
 def option(
@@ -1045,7 +1143,7 @@ def option(
     def f(
         cs: Sequence[str],
     ) -> Result[Parse[Sequence[KeyValue[str]]]]:
-        parser = equals(_flag) >= (lambda _: item(dest, help_name=dest.upper()))
+        parser = matches(_flag) >= (lambda _: item(dest, help_name=dest.upper()))
         return parser.parse(cs)
 
     parser = Parser(f, usage=None, helps={})
